
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "csi_f.h"
#include "rif_r.h"
#include "cbp.h"
#include "sif_f.h"

extern void	Error(char *msg),
		I0Correction(int Nx,int y1,int y2,FOM **I,
			     int Nt,double *T,int Nl,int Nr);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error("multi-threading initialization error.")
#define TERM_MT(T)	\
if (TERM_THREAD(T)) Error("multi-threading termination error.")

#define LEN	2048

typedef struct {
		char	form[LEN];
		int	layer,N;
		FOM	**F;
	} Struct;

static FUNCTION_T	Store(void *a)
{
	Struct	*s=(Struct *)a;
	char	path[LEN];

	(void)sprintf(path,s->form,s->layer);
	StoreImageFile_Float(path,s->N,s->N,s->F,SIF_F_desc);
	(void)printf("%d\t%s\n",s->layer,SIF_F_desc);

	return RETURN_VALUE;
}

#define EPS	1e-9
#define DEG	3.14159265358979323846/180.0

int	main(int argc,char **argv)
{
	char		**path;
	int		Nx,Ny,Nv,Nt,Nl,Nr,rc,Nz,z1,z2,y1,y2,y,z,v,x,len;
	double		Dr,RC,RA0,y0,dy,Dz=0.0,*top=NULL;
	Float		**P,***W,*w,*P_F,**F;
	FOM		**I,*s,*i;
	Struct		S;
	THREAD_T	T;

#ifdef	XR
	if (argc<11 || argc>14)
	    Error(
"usage : rr_tg RR/ NF top left right {layer1 layer2} SDD {-}Dp RC {Dz} RA0 TG/"
	    );
#else
	if (argc<10 || argc>13)
Error("usage : ri_tg RI/ NF top left right {layer1 layer2} Dr RC {Dz} RA0 TG/");
#endif
	path=CheckSliceImages_Float(argv[1],argv[2],&Nx,&Ny,&Nv,NULL);

	if (Nv<1) Error("too few views.");

	free(path[--Nv]);

	if ((Nt=atoi(argv[3]))>=Ny) Error("bad top margin.");

	if ((Nl=atoi(argv[4]))<0) Nl=0;
	if ((Nr=atoi(argv[5]))<0) Nr=0;
	if (Nl+Nr>=Nx) Error("bad side margin.");

#ifdef	XR
#define NO_L12	8

	rc=(argc==11 || argc==12)?NO_L12:NO_L12+2;
{
	double	SDD=atof(argv[rc-2]),
		Dp=atof(argv[rc-1]);

	if (SDD<EPS) Error("bad sample-detector distance.");

	if (-EPS<Dp && Dp<EPS)
	    Error("bad interval of detectors on refraction profile.");

	Dr=SDD/Dp;
}
#else
#define NO_L12	7

	rc=(argc==10 || argc==11)?NO_L12:NO_L12+2;

	if ((Dr=atof(argv[rc-1]))<EPS)
	    Error("bad horizontal interval of detectors on raw image.");
#endif
	if ((RC=atof(argv[rc]))<=0.0 || RC>=(double)(Nx-1))
	    Error("bad rotation center (used as origin of detector array).");

	if (rc+1==argc-2)
	    Nz=Ny;
	else
	    if ((Nz=Ny+(int)ceil(fabs(Dz=atof(argv[rc+1]))))>=2*Ny)
		Error("too large vertical displacement.");

	if (rc==NO_L12) {
	    z1=0; z2=Nz-1;
	}
	else
	    if ((z1=atoi(argv[6]))<0 ||
		(z2=atoi(argv[7]))>=Nz || z1>z2) Error("bad slice range.");

	RA0=atof(argv[argc-2])*DEG;

	P=InitCBP(Nx,Nv);

	if (Dz<0.0) {
	    y0=0.5+Dz; if ((y1=z1+(int)floor(y0))<0) y1=0;
	}
	else {
	    y0=0.5; if ((y1=z1+(int)floor(y0-Dz))<0) y1=0;
	}
        y=(y2=(z2<Ny)?z2:Ny-1)-y1+1;
	if ((W=(Float ***)malloc(sizeof(Float **)*y))==NULL ||
	    (*W=(Float **)malloc(sizeof(Float *)*y*Nv))==NULL ||
	    (**W=(Float *)malloc(sizeof(Float)*y*Nv*Nx))==NULL ||
	    (S.F=(FOM **)malloc(sizeof(FOM *)*Nx))==NULL ||
	    (*(S.F)=(FOM *)malloc(sizeof(FOM)*Nx*Nx))==NULL ||
	    (Nt>0 && (top=(double *)malloc(sizeof(double)*Nx))==NULL))
	    Error("memory allocation error.");

	for (z=0; z<y; z++) {
	    if (z!=0) W[z]=W[z-1]+Nv;

	    for (v=z==0; v<Nv; v++) W[z][v]=W[z][v-1]+Nx;
	}
	for (y=1; y<Nx; y++) S.F[y]=S.F[y-1]+Nx;

	for (v=0; v<Nv; v++) {
	    ReadImageFile_Raw(path[v],&x,&y,&I);
	    if (Nx!=x || Ny!=y) Error("image size not match.");

	    I0Correction(Nx,y1,y2,I,Nt,top,Nl,Nr);

	    for (y=y1; y<=y2; y++) {
		w=W[y-y1][v]; i=I[y]; for (x=0; x<Nx; x++) *(w++)=(*(i++));
	    }
	    for (y=0; y<Ny; y++) free(I[y]); free(I); free(path[v]);
	}
	free(path); if (Nt>0) free(top);

	S.N=Nx;

	len=1; for (z=10; z<Nz; z*=10) ++len;
	(void)sprintf(S.form,"%s/%%0%dd.tif",argv[argc-1],len);

	dy=Dz/(double)Nv; 
	for (z=z1; z<=z2; z++) {
	    for (v=0; v<Nv; v++) {
		P_F=P[v];
		if ((y=z+(int)floor(y0-dy*(double)v))<y1 || y>y2)
		    for (x=0; x<Nx; x++) *(P_F++)=0.0;
		else {
		    w=W[y-y1][v]; for (x=0; x<Nx; x++) *(P_F++)=(*(w++));
		}
	    }
	    F=CBP(Dr,-RC,RA0);

	    if (z!=z1) TERM_MT(T);

	    S.layer=z;

	    for (y=0; y<Nx; y++) {
		s=S.F[y]; P_F=F[y]; for (x=0; x<Nx; x++) *(s++)=(*(P_F++));
            }
	    INIT_MT(T,Store,&S);
	}
	TERM_MT(T);

	free(*(S.F)); free(S.F); free(**W); free(*W); free(W); TermCBP();

	return 0;
}
