
#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* using strlen() */
#include <limits.h>
#include "sif_f.h"

#ifndef	FOF
#define FOF	float
#endif

FOM	SIF_F_min,SIF_F_max;
char	SIF_F_form[SIF_F_LEN]=SIF_F_FORM,
	SIF_F_desc[SIF_F_LEN]="";

#if	UINT_MAX==4294967295
typedef unsigned int	Word4;
#elif	ULONG_MAX==4294967295
typedef unsigned long	Word4;
#else
#error no suitable type for 32-bit integer.
#endif

#if	UINT_MAX==65535
typedef unsigned int	Word2;
#elif	USHRT_MAX==65535
typedef unsigned short	Word2;
#else
#error no suitable type for 16-bit integer.
#endif

static FILE	*file;
static Word4	bytes;

#define PUT1(data)	(void)fputc((int)(data),file)
#define WRITE(ptr,size)	(void)fwrite(ptr,(size_t)(size),1,file)

static void	Put2(Word2 data)
{
	WRITE(&data,2);
}

static void	Put4(Word4 data)
{
	WRITE(&data,4);
}

#ifdef	LZW
static int	LZWnum,LZWlen;
static Word4	LZWbuf;

static void	LZWPut(int code)
{
	LZWnum+=LZWlen; LZWbuf=(LZWbuf<<LZWlen)|code;
	while (LZWnum>=8) {
	    LZWnum-=8; PUT1((LZWbuf>>LZWnum)&0xff); ++bytes;
	}
}

static Word4	LZWstr[9001];
static int	LZWcode,LZWroof,LZWhead;

static void	InitEncode()
{
	int	idx;

	bytes=0; LZWnum=0; LZWlen=9; LZWPut(256);
	for (idx=0; idx<9001; idx++) LZWstr[idx]=0;
	LZWcode=258; LZWroof=512; LZWhead=4096;
}

static void	Encode(int tail)
{
	Word4	h_t;
	int	idx,stp;

	if (LZWhead!=4096) {
	    h_t=(((Word4)LZWhead<<8)|tail)<<12;
	    if (LZWstr[idx=LZWhead^(tail<<5)]!=0) {
		if ((LZWstr[idx]&0xfffff000)==h_t) {
		    LZWhead=(int)(LZWstr[idx]&0xfff); return;
		}
		stp=(idx==0)?9000:idx;
		while (LZWstr[idx=(idx+stp)%9001]!=0)
		    if ((LZWstr[idx]&0xfffff000)==h_t) {
			LZWhead=(int)(LZWstr[idx]&0xfff); return;
		    }
	    }
	    LZWPut(LZWhead);
	    if (LZWcode!=4096-1-1) {
		LZWstr[idx]=h_t|LZWcode;
		if (++LZWcode==LZWroof) {
		    ++LZWlen; LZWroof<<=1;
		}
	    }
	    else {
		LZWPut(256); LZWlen=9;
		for (idx=0; idx<9001; idx++) LZWstr[idx]=0;
		LZWcode=258; LZWroof=512;
	    }
	}
	LZWhead=tail;
}

static Word4	TermEncode()
{
	if (LZWhead!=4096) {
	    LZWPut(LZWhead); if (++LZWcode==LZWroof) ++LZWlen;
	}
	LZWPut(257);
	if (LZWnum>0) {
	    PUT1((LZWbuf<<(8-LZWnum))&0xff); ++bytes;
	}
	return bytes;
}
#else
#define InitEncode()	bytes=0

static void	Encode(int code)
{
	PUT1(code); ++bytes;
}

#define TermEncode()	bytes
#endif

#define PutIFD(tag,type,count)	Put2(tag); Put2(type); Put4(count)
#define PutIFD3(tag,data)	PutIFD(tag,3,1); Put2(data); Put2(0)
#define PutIFD4(tag,count,data)	PutIFD(tag,4,count); Put4(data)

void	StoreImageFile_Float(char *path,int Nx,int Ny,FOM **cell,char *desc)
{
	Word4		*cnt,ifd,
			ofs=2+2+4,
			bpr=(Word4)Nx*sizeof(FOF);
	Word2		mn=0x002a;
	int		y,x,len,rem,idx,
			rps=(bpr>8192)?1:(int)(8192/bpr),
			spi=(Ny+rps-1)/rps;
	FOF		fof;
	unsigned char	*byte=(unsigned char *)&fof;

	if ((cnt=(Word4 *)malloc((size_t)spi*4))==NULL) {
	    (void)fprintf(stderr,"%s : no allocatable memory.\n",path); exit(1);
	}
	if ((file=fopen(path,"wb"))==NULL) {
	    (void)fprintf(stderr,"%s : file not stored.\n",path); exit(1);
	}
	WRITE((*((unsigned char *)&mn)==mn)?"II":"MM",2);
	Put2(mn);

	Put4(0);	/* tentative IFD offset (at file offset 4) */

	SIF_F_min=SIF_F_max=cell[0][0];
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[y][x]<SIF_F_min) SIF_F_min=cell[y][x]; else
	    if (cell[y][x]<SIF_F_max) SIF_F_max=cell[y][x];

	if (desc==NULL)
	    len=0;
	else {
	    if (desc==SIF_F_desc)
		(void)sprintf(SIF_F_desc,SIF_F_form,(double)SIF_F_min,
						    (double)SIF_F_max);

	    len=strlen(desc)+1;
	}
	if (len>4) {
	    WRITE(desc,len);
	    if ((ofs+=len)&1) {
		PUT1(0); ++ofs;
	    }
	}
	rem=0;
	for (y=0; y<Ny; y++) {
	    if (rem==0) InitEncode();

	    for (x=0; x<Nx; x++) {
		fof=cell[y][x];
		for (idx=0; idx<sizeof(FOF); idx++) Encode(byte[idx]);
	    }
	    if ((rem=y+1)==Ny || (rem%=rps)==0)
		if ((cnt[y/rps]=TermEncode())&1) PUT1(0);
	}
	if (spi==1)
	    ifd=ofs+(*cnt)+(*cnt&1);
	else {
	    for (idx=0; idx<spi; idx++) {
		Put4(ofs); if ((ofs+=cnt[idx])&1) ++ofs;
	    }
	    WRITE(cnt,spi*4); ifd=ofs+spi*4*2;
	}
	if (sizeof(FOF)==8) {
	    fof=SIF_F_min; WRITE(&fof,8);
	    fof=SIF_F_max; WRITE(&fof,8); ifd+=(8*2);
	}
	Put2((len)?14:13);		/* number of tags */

	if (Nx>0xffff) {
	    PutIFD4(0x100,1,Nx);	/* 0: image width */
	}
	else {
	    PutIFD3(0x100,(Word2)Nx);	/* 0: image width */
	}
	if (Ny>0xffff) {
	    PutIFD4(0x101,1,Ny);	/* 1: image length */
	}
	else {
	    PutIFD3(0x101,(Word2)Ny);	/* 1: image length */
	}
	PutIFD3(0x102,(Word2)(sizeof(FOF)*8));
					/* 2: bits per sample */
#ifdef	LZW
	PutIFD3(0x103,5);		/* 3: image compression */
#else
	PutIFD3(0x103,1);		/* 3: image compression */
#endif
	PutIFD3(0x106,1);		/* 4: photometric interpretation */
	if (len) {
	    PutIFD(0x10e,2,len);
	    if (len<=4) {
		WRITE(desc,len); while (++len<=4) PUT1(0);
	    }
	    else
		Put4(8);		/* 5: image description */
	}
	PutIFD4(0x111,spi,ofs);		/* 5/6: strip offset */
	PutIFD3(0x115,1);		/* 6/7: samples per pixel */
	PutIFD4(0x116,1,rps);		/* 7/8: rows per strip */
	PutIFD4(0x117,spi,(spi==1)?*cnt:ofs+spi*4);
					/* 8/9: strip byte counts */
	PutIFD3(0x11c,1);		/* 9/10: planar configuration */
	PutIFD3(0x153,3);		/* 10/11: sample format */
	if (sizeof(FOF)==4) {
	    PutIFD(0x154,11,1); fof=SIF_F_min; WRITE(&fof,4);
					/* 11/12: minimum sample value */
	    PutIFD(0x155,11,1); fof=SIF_F_max; WRITE(&fof,4);
					/* 12/13: maximum sample value */
	}
	else {
	    PutIFD(0x154,12,1); Put4(ifd-8*2);
					/* 11/12: minimum sample value */
	    PutIFD(0x155,12,1); Put4(ifd-8);
					/* 12/13: maximum sample value */
	}
	Put4(0);			/* end of tags */

	(void)fseek(file,4L,0); Put4(ifd);

	(void)fclose(file);

	free(cnt);
}
