
#include <stdio.h>
#include <stdlib.h>
#include "sif_f.h"

extern void	Error(char *msg);

static unsigned	Image=0;

#define LEN	2048

static void	Abort(char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s at image %u.",msg,Image); Error(str);
}

static void	Skip(FILE *file,size_t size,char *msg)
{
	char	str[LEN];

	if (file!=stdin)
	    size=(size_t)fseeko(file,(off_t)size,SEEK_CUR);
	else
	    do if (fgetc(file)==EOF) break; while (--size!=0);

	if (size!=0) {
	    (void)sprintf(str,"%s skip error",msg); Abort(str);
	}
}

#define SHORT	unsigned short

static unsigned	Np,Nz,Npz,NumOfImages=0;
static SHORT	RemChannels=1;
static size_t	SNpz;

static void	Header(FILE *file,char *msg)
{
	char		str[LEN],header[64];
	unsigned	*NOI=(unsigned *)(header+14);
	SHORT		*CL=(SHORT *)(header+2),
			*IW=(SHORT *)(header+4),
			*IH=(SHORT *)(header+6),
			*FT=(SHORT *)(header+12),
			*RC=(SHORT *)(header+18);

	if (feof(file) || (NumOfImages==0 && RemChannels==0)) {
	    (void)sprintf(str,"missing %s",msg); Abort(str);
	}
	if (fread(&header,64,1,file)!=1 ||
	    header[0]!='I' || header[1]!='M') Abort("header read error");

	if (*FT!=2) Abort("unsupported file type");

	if (Image==0) {
	    if (*IW==0 || *IH==0) Abort("bad size");

	    SNpz=sizeof(SHORT)*(Npz=(Np=(*IW))*(Nz=(*IH)));
	}
	else
	    if (*IW!=Np || *IH!=Nz) Abort("size not match");

	if (*CL!=0) Skip(file,*CL,"comment");

	if (NumOfImages==0) {
	    NumOfImages=(*NOI); RemChannels=(*RC);
	}
	--NumOfImages;
}

static void	Read(FILE *file,SHORT *data,char *msg)
{
	char	str[LEN];

	if (fread(data,SNpz,1,file)!=1) {
	    (void)sprintf(str,"%s read error",msg); Abort(str);
	}
}

int	main(int argc,char **argv)
{
	FILE		*HIS;
	int		Ns,Nv,Nd,p,z,s,l,v,Wm=(-1),Ws=0;
	double		*K,*D;
	FOM		**RI,**RR;
	SHORT		*I;
	unsigned	pz;
	char		FI[LEN],FR[LEN],path[LEN];

	if (argc!=7 && argc!=9 && argc!=10 && argc!=12) Error(
	"usage : his2raw HIS scans views darks {Ka Kb Kc} {Wm Ws} RI/ RR/\n"
	"        his2raw HIS scans -view darks {Ka Kb Kc} {Wm Ws} RI.tif RR.tif"
	);
	if (argv[1][0]=='-')
	    HIS=stdin;
	else
	    if ((HIS=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if ((Ns=atoi(argv[2]))<2) Error("bad number of scans."); --Ns;

	if (argv[3][0]=='-')
	    Nv=atoi(argv[3]);
	else
	    if ((Nv=atoi(argv[3]))<1) Error("bad number of views.");

	if ((Nd=atoi(argv[4]))<1) Error("bad number of darks.");

	if (argc==9 || argc==12) {
	    if ((Wm=atoi(argv[argc-4]))<0) Error("bad mountain width.");
	    if ((Ws=atoi(argv[argc-3]))<2) Error("bad sampling width.");
	}
	for (Image=0; Image<Nd; Image++) {
	    Header(HIS,"dark data");

	    if (Image==0) {
		if ((K=(double *)malloc(sizeof(double)*Np))==NULL ||
		    (D=(double *)malloc(sizeof(double)*Npz))==NULL ||
		    (RI=(FOM **)malloc(sizeof(FOM *)*Nz))==NULL ||
		    (*RI=(FOM *)malloc(sizeof(FOM)*Nz*Ns))==NULL ||
		    (RR=(FOM **)malloc(sizeof(FOM *)*Nz))==NULL ||
		    (*RR=(FOM *)malloc(sizeof(FOM)*Nz*Ns))==NULL ||
		    (I=(SHORT *)malloc(SNpz))==NULL)
		    Error("memory allocation error.");

		if (argc==7 || argc==9)
		    for (p=0; p<Np; p++) K[p]=1.0;
		else
{
	double	a=atof(argv[5]),b=atof(argv[6]),c=atof(argv[7]),k=0.0;

	for (p=0; p<Np; p++) {
	    if ((K[p]=a*(double)p*(double)p+b*(double)p+c)<=0.0)
		Error("bad coefficients for vignetting correction.");

	    if (k<K[p]) k=K[p];
	}
	for (p=0; p<Np; p++) K[p]=k/K[p];
}
		for (pz=0; pz<Npz; pz++) D[pz]=0.0;

		for (z=1; z<Nz; z++) {
		    RI[z]=RI[z-1]+Ns; RR[z]=RR[z-1]+Ns;
		}
	    }
	    Read(HIS,I,"dark data");

	    for (pz=0; pz<Npz; pz++) D[pz]+=(double)I[pz];
	}
	for (pz=0; pz<Npz; pz++) D[pz]/=(double)Nd;

	if (Nv<=0) {
	    while (++Nv<1)
		for (s=0; s<=Ns; s++, Image++) {
		    Header(HIS,"unused data"); Skip(HIS,SNpz,"unused data");
		}

	    for (l=0; (FI[l]=argv[argc-2][l])!='\0'; l++) ;
	    for (l=0; (FR[l]=argv[argc-1][l])!='\0'; l++) ; l=0;
	}
	else {
	    l=1; for (v=10; v<Nv; v*=10) ++l;

	    (void)sprintf(FI,"%s/%%0%dd.tif",argv[argc-2],l);
	    (void)sprintf(FR,"%s/%%0%dd.tif",argv[argc-1],l);
	}
	for (v=0; v<Nv; v++) {
	    Header(HIS,"junk data"); Skip(HIS,SNpz,"junk data"); ++Image;

	    for (s=0; s<Ns; s++) {
		Header(HIS,"data"); Read(HIS,I,"data"); ++Image;

		for (pz=0, z=0; z<Nz; z++, pz+=Np)
{
	int	p0,p1,p2,p3,p4,c=0,num=0;
	double	i_d,top,det,a,b,
		*d=D+pz,x=0.0,y=0.0,xx=0.0,xy=0.0,sum=0.0,mom=0.0;
	SHORT	*i=I+pz;

	for (p=0; p<Np; p++) {
	    i_d=K[p]*((double)i[p]-d[p]);
	    if (p==0 || i_d>top) {
		p0=p; top=i_d;
	    }
	}
	p1=(p2=p0-Wm)-Ws;
	p4=(p3=p0+Wm)+Ws;

	for (p=0; p<Np; p++)
	    if (p<p2 || p>p3) {
		i_d=K[p]*((double)i[p]-d[p]);
		if (p>=p1 && p<=p4) {
		    ++c; x+=(double)p; y+=i_d;
			 xx+=((double)p*(double)p); xy+=((double)p*i_d);
		}
		++num; sum+=i_d; mom+=((double)p*i_d);
	    }

	if (c>=2) {
	    det=xx*(double)c-x*x; a=((double)c*xy-x*y)/det; b=(xx*y-x*xy)/det;
	    for (p=(p2<0)?0:p2; p<=p3 && p<Np; p++) {
		++num; i_d=a*(double)p+b; sum+=i_d; mom+=((double)p*i_d);
	    }
	}
	if (num==0 || sum<=0.0)
	    RI[z][s]=RR[z][s]=0.0;
	else {
	    RI[z][s]=sum/(double)num; RR[z][s]=mom/sum;
	}
}
	    }
	    if (l) (void)printf("%d\t",v);

	    (void)sprintf(path,FI,v);
	    StoreImageFile_Float(path,Ns,Nz,RI,SIF_F_desc);
	    (void)printf("%s\t",SIF_F_desc);

	    (void)sprintf(path,FR,v);
	    StoreImageFile_Float(path,Ns,Nz,RR,SIF_F_desc);
	    (void)printf("%s\n",SIF_F_desc);
	}
	free(I); free(*RR); free(RR); free(*RI); free(RI); free(D); free(K);

	if (HIS!=stdin) (void)fclose(HIS);

	return 0;
}
