
#include <stdio.h>
#include <stdlib.h>
#include "sif_f.h"

extern void	Error(char *msg);

static unsigned	Image=0;

#define LEN	2048

static void	Abort(char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s at image %u.",msg,Image); Error(str);
}

static void	Skip(FILE *file,size_t size,char *msg)
{
	char	str[LEN];

	if (file!=stdin)
	    size=(size_t)fseeko(file,(off_t)size,SEEK_CUR);
	else
	    do if (fgetc(file)==EOF) break; while (--size!=0);

	if (size!=0) {
	    (void)sprintf(str,"%s skip error",msg); Abort(str);
	}
}

#define SHORT	unsigned short

static unsigned	Np,Nz,Npz,NumOfImages=0;
static SHORT	RemChannels=1;
static size_t	SNpz;

static void	Header(FILE *file,char *msg)
{
	char		str[LEN],header[64];
	unsigned	*NOI=(unsigned *)(header+14);
	SHORT		*CL=(SHORT *)(header+2),
			*IW=(SHORT *)(header+4),
			*IH=(SHORT *)(header+6),
			*FT=(SHORT *)(header+12),
			*RC=(SHORT *)(header+18);

	if (feof(file) || (NumOfImages==0 && RemChannels==0)) {
	    (void)sprintf(str,"missing %s",msg); Abort(str);
	}
	if (fread(&header,64,1,file)!=1 ||
	    header[0]!='I' || header[1]!='M') Abort("header read error");

	if (*FT!=2) Abort("unsupported file type");

	if (Image==0) {
	    if (*IW==0 || *IH==0) Abort("bad size");

	    SNpz=sizeof(SHORT)*(Npz=(Np=(*IW))*(Nz=(*IH)));
	}
	else
	    if (*IW!=Np || *IH!=Nz) Abort("size not match");

	if (*CL!=0) Skip(file,*CL,"comment");

	if (NumOfImages==0) {
	    NumOfImages=(*NOI); RemChannels=(*RC);
	}
	--NumOfImages;
}

static void	Read(FILE *file,SHORT *data,char *msg)
{
	char	str[LEN];

	if (fread(data,SNpz,1,file)!=1) {
	    (void)sprintf(str,"%s read error",msg); Abort(str);
	}
}

int	main(int argc,char **argv)
{
	FILE		*HIS;
	int		z,Nd=0;
	unsigned	RP,pz;
	double		*D;
	FOM		**F,*f;
	SHORT		*I;

	if (argc!=4 && argc!=5)
	    Error("usage : his2rp HIS {darks} RP_number RP.tif");

	if (argv[1][0]=='-')
	    HIS=stdin;
	else
	    if ((HIS=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (argc==5 && (Nd=atoi(argv[2]))<1) Error("bad number of darks.");

	if (sscanf(argv[argc-2],"%u",&RP)!=1 || RP<Nd)
	    Error("bad refraction profile number.");

	for (Image=0; Image<=RP; Image++) {
	    Header(HIS,(Image<Nd)?"dark data":"data");

	    if (Image==0) {
		if ((D=(double *)malloc(sizeof(double)*Npz))==NULL ||
		    (F=(FOM **)malloc(sizeof(FOM *)*Nz))==NULL ||
		    (I=(SHORT *)malloc(SNpz))==NULL)
		    Error("memory allocation error.");

		for (pz=0; pz<Npz; pz++) D[pz]=0.0;

		*F=f=(FOM *)D; for (z=1; z<Nz; z++) F[z]=F[z-1]+Np;
	    }
	    if (Image<Nd) {
		Read(HIS,I,"dark data");

		for (pz=0; pz<Npz; pz++) D[pz]+=(double)I[pz];
	    }
	    else if (Image<RP)
		Skip(HIS,SNpz,"data");
	    else
		Read(HIS,I,"data");
	}
	for (pz=0; pz<Npz; pz++) f[pz]=(double)I[pz]-D[pz]/(double)Nd;

	StoreImageFile_Float(argv[argc-1],Np,Nz,F,SIF_F_desc);
	(void)puts(SIF_F_desc);

	free(I); free(F); free(D);

	if (HIS!=stdin) (void)fclose(HIS);

	return 0;
}
