
#include <stdio.h>
#include <stdlib.h>
#include "sif_f.h"

extern void	Error(char *msg);

static unsigned	Image=0;

#define LEN	2048

static void	Abort(char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s at image %u.",msg,Image); Error(str);
}

static void	Skip(FILE *file,size_t size,char *msg)
{
	char	str[LEN];

	if (file!=stdin) {
	    if (fseeko(file,(off_t)size,SEEK_CUR)) goto abort;
	}
	else
	    do if (fgetc(file)==EOF) goto abort; while (--size!=0);

	return;
abort:;
	(void)sprintf(str,"%s skip error",msg); Abort(str);
}

#define SHORT	unsigned short

static unsigned	NumOfImages=0;
static SHORT	RemChannels=1;
static size_t	SNpz,Npz;
static int	Np,Nz;
static double	ts;

static void	Header(FILE *file,char *msg)
{
	char		str[LEN],header[64];
	unsigned	*NOI=(unsigned *)(header+14);
	SHORT		*CL=(SHORT *)(header+2),
			*IW=(SHORT *)(header+4),
			*IH=(SHORT *)(header+6),
			*FT=(SHORT *)(header+12),
			*RC=(SHORT *)(header+18);
	double		*TS=(double *)(header+22);

	if (feof(file) || (NumOfImages==0 && RemChannels==0)) {
	    (void)sprintf(str,"missing %s",msg); Abort(str);
	}
	if (fread(&header,64,1,file)!=1 ||
	    header[0]!='I' || header[1]!='M') Abort("header read error");

	if (*FT!=2) Abort("unsupported file type");

	if (Image==0) {
	    if (*IW==0 || *IH==0) Abort("bad size");

	    SNpz=sizeof(SHORT)*(Npz=(Np=(*IW))*(Nz=(*IH)));
	}
	else
	    if (*IW!=Np || *IH!=Nz) Abort("size not match");

	if (*CL!=0) Skip(file,*CL,"comment");

	if (NumOfImages==0) {
	    NumOfImages=(*NOI); RemChannels=(*RC);
	}
	--NumOfImages;

	ts=(*TS);
}

static void	Read(FILE *file,SHORT *data,char *msg)
{
	char	str[LEN];

	if (fread(data,SNpz,1,file)!=1) {
	    (void)sprintf(str,"%s read error",msg); Abort(str);
	}
}

int	main(int argc,char **argv)
{
	FILE	*HIS;
	int	Ns,Nv,Nd,Wm,Ws,Ns1,z1,z2,v,s,z,p,p0,p1,p2,p3,p4,c,num;
	SHORT	*I,*i;
	double	*D,*d,mean,i_d,top,x,xx,y,xy,sum,det,a,b;
	FOM	**Time,**Mean;
	size_t	pz;

	if (argc!=8 && argc!=9 && argc!=10 && argc!=11)
Error("usage : his2tm HIS scans views darks Wm Ws {z1 z2} {time.tif} mean.tif");

	if (argv[1][0]=='-')
	    HIS=stdin;
	else
	    if ((HIS=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if ((Ns=atoi(argv[2]))<2) Error("bad number of scans.");
	if ((Nv=atoi(argv[3]))<1) Error("bad number of views.");
	if ((Nd=atoi(argv[4]))<1) Error("bad number of darks.");

	Wm=atoi(argv[5]);
	Ws=atoi(argv[6]);

	Ns1=Ns-1;

	NumOfImages=0;
	RemChannels=1;
	for (Image=0; Image<Nd; Image++) {
	    Header(HIS,"dark data");

	    if (Image==0) {
		if (argc==8 || argc==9) {
		    z1=0; z2=Nz-1;
		}
		else
		    if ((z1=atoi(argv[7]))<0 ||
			(z2=atoi(argv[8]))>=Nz || z1>z2)
			Error("bad slice range.");

		if ((I=(SHORT *)malloc(SNpz))==NULL ||
		    (D=(double *)malloc(sizeof(double)*Npz))==NULL ||
		    (Time=(FOM **)malloc(sizeof(FOM *)*Nv))==NULL ||
		    (*Time=(FOM *)malloc(sizeof(FOM)*Nv*Ns1))==NULL ||
		    (Mean=(FOM **)malloc(sizeof(FOM *)*Nv))==NULL ||
		    (*Mean=(FOM *)malloc(sizeof(FOM)*Nv*Ns1))==NULL)
		    Error("memory allocation error.");

		d=D; for (pz=0; pz<Npz; pz++) *(d++)=0.0;

		for (v=1; v<Nv; v++) {
		    Time[v]=Time[v-1]+Ns1; Mean[v]=Mean[v-1]+Ns1;
		}
	    }
	    Read(HIS,i=I,"dark data");

	    d=D; for (pz=0; pz<Npz; pz++) *(d++)+=((double)*(i++)/(double)Nd);
	}
	for (v=0; v<Nv; v++) {
	    Header(HIS,"junk data"); Skip(HIS,SNpz,"junk data"); ++Image;

	    for (s=0; s<Ns1; s++) {
		Header(HIS,"data"); Read(HIS,I,"data"); ++Image;

		Time[v][s]=ts/1000.0;

		mean=0.0;
		for (z=z1; z<=z2; z++) {
		    pz=(size_t)Np*z; d=D+pz; i=I+pz;
		    for (p=0; p<Np; p++) {
			i_d=(double)i[p]-d[p];
			if (p==0 || i_d>top) {
			    p0=p; top=i_d;
			}
		    }
		    p1=(p2=p0-Wm)-Ws;
		    p4=(p3=p0+Wm)+Ws; c=num=0; x=xx=y=xy=sum=0.0;
		    for (p=0; p<Np; p++)
			if (p<p2 || p>p3) {
			    i_d=(double)i[p]-d[p];
			    if (p>=p1 && p<=p4) {
				++c; x+=(double)p; xx+=((double)p*(double)p);
				     y+=i_d; xy+=((double)p*i_d);
			    }
			    ++num; sum+=i_d;
			}

		    if (c>=2) {
			det=xx*(double)c-x*x; a=((double)c*xy-x*y)/det;
					      b=(xx*y-x*xy)/det;
			for (p=(p2<0)?0:p2; p<=p3 && p<Np; p++) {
			    ++num; sum+=(a*(double)p+b);
			}
		    }
		    if (num>0 && sum>0.0) mean+=(sum/(double)num);
		}
		Mean[v][s]=mean/(double)(z2-z1+1);
	    }
	}
	if (argc==9 || argc==11)
	StoreImageFile_Float(argv[argc-2],Ns1,Nv,Time,SIF_F_desc);

	StoreImageFile_Float(argv[argc-1],Ns1,Nv,Mean,SIF_F_desc);

	free(*Mean); free(Mean); free(*Time); free(Time); free(D); free(I);

	if (HIS!=stdin) (void)fclose(HIS);

	return 0;
}
