#!/bin/csh
# usage :
# pile_[h,v].csh TG/ images_base step number rows cols {CTV_base step} TIFF
#
set ratio=(0.075 0.05 0.025)	# LU and RD labels, and space between images
@ pv = 255			# PV for LU and RD labels
#
set tmp=/tmp/pile.$$
#
alias STDERR 'echo echo 1\>\&2 "\!*" | /bin/sh'
#
set csh=`basename $0`
switch($csh)
  case pile_v.csh :
    set h_v= ; breaksw
  case pile_h.csh :
    set h_v=- ; breaksw
  default :
    STDERR $csh : unknown C-shell script. ; exit 1
endsw
#
if ($#argv != 7 && $#argv != 9) then
  STDERR usage : $csh TG/ images_base step number cols rows {CTV_base step} TIFF
  exit 1
endif
#
if (!(-d $argv[1])) then
  STDERR $argv[1] : directory not found. ; exit 1
endif
#
if ($#argv == 7) then
  set b_s=
else
  if (`echo $argv[8] | awk '{ print ($1<=0) }'`) then
    STDERR $argv[8] : bad CT-value step. ; exit 1
  endif
#
  set b_s=($argv[7-8])
endif
#
set log=(`( echo $ratio ; echo $argv ; ls $argv[1] ) | \
          awk 'NR==1 { LU=$1; RD=$2; II=$3 } \
               NR==2 { TG=$1; B=$2; S=$3; N=($4<(n=$5*$6))?$4:n } \
               NR>=3 && (n=NR-3-B)>=0 && n%S==0 && n/S<N { \
                 (p="pig " (f=TG "/" (t=$0))) | getline; close(p); \
                 if (NF<4) exit 1; \
                 if ($3!=1) { \
                   print f,": bad image type." > "/dev/stderr"; exit 1 } \
                 CSV[C++]=t "," $4; if (C==1 || X>$1) X=$1; \
                                    if (C==1 || Y>$2) Y=$2 } \
               END { z=(X+Y)/2; print LU*z,RD*z,int(II*z+1); \
                     for (c=0; c<C; c++) print CSV[c] }'`)
if ($status || $#log < 4) exit 1
#
mkdir $tmp
foreach csv ($log[4-])
  set ssv=(`echo $csv | tr , ' '`)
  t2t_float ${argv[1]}/$ssv[1] $b_s 8 ${tmp}/$ssv[1] >& /dev/null
#
  ( echo Helvetica $log[1] $pv LU `echo $ssv[1] | sed s/\\.tif\$//g` ; \
    pid ${argv[1]}/$ssv[1] | \
    awk -v BPS=$ssv[2] 'NF==3 { print $2,"~",$3*(2^BPS-1) } \
                        NF==2 || NF==4 { print $1,"~",$2 } \
                        NF==6 || NF==8 { print $5,"~",$6 }' | \
    sed "s/^/Helvetica $log[2] $pv RD /g" ) | add_label ${tmp}/$ssv[1]
end
pile_gray ${h_v}$argv[5-6] -$log[3] -$log[3] 255 $argv[$#argv] ${tmp}/*
rm -rf $tmp
