
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "csi_f.h"
#include "rif_r.h"
#include "sif_f.h"

extern void	Error(char *msg),
		I0Correction(int Nx,int y1,int y2,FOM **I,
			     int Nt,double *T,int Nl,int Nr);

#if	defined(XR)
#define ORG	"RR/"
#define RUN	"rr2mr"
#define NEW	"MR.tif"
#else
#define ORG	"RI/"
#define RUN	"ri2mp"
#define NEW	"MP.tif"
#endif

int	main(int argc,char **argv)
{
	char	**path;
	int	Nx,Ny,Nz,Nt,Nl,Nr,y,z,x,y0,Y,ny,d,y1,y2,D;
	double	s,a,s2,b,s1,rmv,RMV,*T=NULL;
	FOM	**M,**I,m;
	FILE	*file;

	if (argc!=7 && argc!=8 && argc!=10)
Error("usage : " RUN " " ORG " NF top left right " NEW " {{y1 y2} SVC.txt}");

	path=CheckSliceImages_Float(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if ((Nt=atoi(argv[3]))>=Ny) Error("bad top margin.");

	if ((Nl=atoi(argv[4]))<0) Nl=0;
	if ((Nr=atoi(argv[5]))<0) Nr=0;
	if (Nl+Nr>=Nx) Error("bad side margin.");

	if ((Nt>0 && (T=(double *)malloc(sizeof(double)*Nx))==NULL) ||
	    (M=(FOM **)malloc(sizeof(FOM *)*Ny))==NULL ||
	    (*M=(FOM *)malloc(sizeof(FOM)*Ny*Nz))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<Ny; y++) M[y]=M[y-1]+Nz;

	for (z=0; z<Nz; z++) {
	    ReadImageFile_Raw(path[z],&x,&y,&I);
	    if (Nx!=x || Ny!=y) Error("image size not match.");

	    I0Correction(Nx,0,Ny-1,I,Nt,T,Nl,Nr);

	    for (y=0; y<Ny; y++) {
		s=0.0; for (x=0; x<Nx; x++) s+=I[y][x]; M[y][z]=s/(double)Nx;

		free(I[y]);
	    }
	    free(I); free(path[z]);
	}
	StoreImageFile_Float(argv[6],Nz,Ny,M,SIF_F_desc);
	(void)puts(SIF_F_desc);

	free(path); if (Nt>0) free(T);

	if (argc==8 || argc==10) {
	    if (Nz<2) Error("too few views.");

	    if (argc==8) {
		y0=0; Y=ny=Ny;
	    }
	    else
		if ((y0=atoi(argv[7]))<0 ||
		    (Y=atoi(argv[8])+1)>Ny ||
		    (ny=Y-y0)<=0) Error("bad slice range.");

	    if (argv[argc-1][0]=='-')
		file=stdout;
	    else
	    if ((file=fopen(argv[argc-1],"w"))==NULL) Error("file not open.");

	    for (d=1-ny; d<ny; d++) {
		if (d<0) {
		    y1=y0; y2=Y+d;
		}
		else {
		    y1=y0+d; y2=Y;
		}
		a=(double)d/(double)(Nz-1); s2=0.0;
		for (y=y1; y<y2; y++) {
		    b=(double)y+0.5; s1=0.0;
		    for (z=0; z<Nz; z++) {
			m=M[(int)floor(b-a*(double)z)][z]; s1+=m; s2+=(m*m);
		    }
		    s2-=(s1*s1/(double)Nz);
		}
		(void)fprintf(file,"%d\t%le\n",
				   d,rmv=sqrt(s2/(double)Nz/(double)(y2-y1)));

		if (d==1-ny || rmv<RMV) {
		    D=d; RMV=rmv;
		}
	    }
	    if (file!=stdout) (void)fclose(file);

	    (void)printf("%d\t%le\n",D,RMV);
	}
	free(*M); free(M);

	return 0;
}
