
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "csi_f.h"
#include "rif_r.h"
#include "sif_f.h"

extern void	Error(char *msg),
		I0Correction(int Nx,int y1,int y2,FOM **I,
			     int Nt,double *T,int Nl,int Nr);

static int	Range(int *Ng,int *Nh,int *D1,int *D2,int *Og,int *Oh)
{
	int	g,h,N,L;

	if ((g=(*Nh)+(*D2))>(*Ng)) g=(*Ng);
	if ((h=(*Ng)-(*D1))>(*Nh)) h=(*Nh);

	if (*D2<0) {
	    *Og=0; *Oh=(-(*D2)); *D1-=(*D2); *D2=0; N=1-(*D1);
	}
	else if (*D1>0) {
	    *Og=(*D1); *Oh=0; *D2-=(*D1); *D1=0; N=(*D2)+1;
	}
	else {
	    *Og=(*Oh)=0; N=(-(*D1)>(*D2))?1-(*D1):*D2+1;
	}
	N+=(((*Ng=g-(*Og))>(*Nh=h-(*Oh)))?*Ng:*Nh);

	for (L=1; L<N; L<<=1) ; return L;
}

typedef struct {
		double	r,i;
	} Complex;

static void	FFT(int S,int L,Complex *E,Complex *C)
{
	int	j,k,m,n,p,q,L2=L>>1;
	Complex	c,*e,*a,*b;
	double	r,i,d;

	for (j=L-1, k=L-2; k>0; k--) {
	    for (m=L2; (j^=m)&m; m>>=1) ;

	    if (j>k) {
		c=C[j]; C[j]=C[k]; C[k]=c;
	    }
	}
	for (j=2, k=1, m=L2, n=L; n>1; n=m, m>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=n) {
		r=e->r;
		i=e->i*(double)S;
		for (a=C+p, q=0; q<m; q++, a+=j) {
		    c=(*(b=a+k)); b->r=a->r-(d=c.r*r-c.i*i); a->r+=d;
				  b->i=a->i-(d=c.r*i+c.i*r); a->i+=d;
		}
	    }
}

static Complex	Product(double r1,double i1,double r2,double i2)
{
	Complex	c;

	c.r=r1*r2-i1*i2;
	c.i=r1*i2+i1*r2; return c;
}

#ifndef	M_PI
#define	M_PI	3.14159265358979323846
#endif

FOM	*RMSD(int Ngx,int Ngy,FOM **G,
	      int Nhx,int Nhy,FOM **H,int Dx1,int Dy1,int Dx2,int Dy2)
{
	int	Lx,Ogx,Ohx,Lx2,x,Lx1,Ngx1,Ngx2,Nhx1,Nhx2,gx,hx,
		Ly,Ogy,Ohy,Ly2,y,Ly1,Ngy1,Ngy2,Nhy1,Nhy2,gy,hy,Lxx,z;
	FOM	*P,**Gy,**Hy,*Gx,*Hx,*p;
	double	d,gr,gi,hr,hi,t,r,i,Lxy,Lxyy;
	Complex	*Ex,*Ey,**Q,*U,*V,*ex,*ey,*q,*u,*v,g,h,c;

	Lx=Range(&Ngx,&Nhx,&Dx1,&Dx2,&Ogx,&Ohx); Lx2=Lx*2;
	Ly=Range(&Ngy,&Nhy,&Dy1,&Dy2,&Ogy,&Ohy); Ly2=Ly*2;

	if ((P=(FOM *)malloc(sizeof(FOM)*(Dy2-Dy1+1)*(Dx2-Dx1+1)))==NULL ||
	    (Ex=(Complex *)malloc(sizeof(Complex)*Lx2))==NULL ||
	    (Ey=(Complex *)malloc(sizeof(Complex)*Ly2))==NULL ||
	    (Q=(Complex **)malloc(sizeof(Complex *)*Ly))==NULL ||
	    (*Q=(Complex *)malloc(sizeof(Complex)*Ly*Lx))==NULL ||
	    (U=(Complex *)malloc(sizeof(Complex)*Ly))==NULL ||
	    (V=(Complex *)malloc(sizeof(Complex)*Ly))==NULL) return NULL;

	for (ex=Ex, x=0; x<Lx; x++, ex++) {
	    d=M_PI*(double)x/(double)Lx; ex[Lx].r=(-(ex->r=cos(d)));
					 ex[Lx].i=(-(ex->i=sin(d)));
	}
	for (ey=Ey, y=0; y<Ly; y++, ey++) {
	    d=M_PI*(double)y/(double)Ly; ey[Ly].r=(-(ey->r=cos(d)));
					 ey[Ly].i=(-(ey->i=sin(d)));

	    if (y>0) Q[y]=Q[y-1]+Lx;
	}
	Lx1=Lx-1; Ngx1=Ngx-1; Nhx1=Nhx-1; Ngx2=Ngx%2; Nhx2=Nhx%2;
	Ly1=Ly-1; Ngy1=Ngy-1; Nhy1=Nhy-1; Ngy2=Ngy%2; Nhy2=Nhy%2;
/*
    ----
*/
	for (Gy=G+Ogy, Hy=H+Ohy, y=0; y<Ly; y++) {
	    x=0; q=Q[y];
	    if (y<Ngy)
		for (Gx=(*(Gy++))+Ogx; x<Ngx; x++) {
		    d=(*(Gx++)); (q++)->r=d*d;
		}

	    for (; x<Lx; x++) (q++)->r=0.0;

	    x=0; q=Q[y];
	    if (y<Nhy)
		for (Hx=(*(Hy++))+Ohx; x<Nhx; x++) {
		    d=(*(Hx++)); (q++)->i=d*d;
		}

	    for (; x<Lx; x++) (q++)->i=0.0;

	    if (y<Ngy || y<Nhy) FFT(-1,Lx,Ex,Q[y]);
	}
/*
    x=0
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][0];

	FFT(-1,Ly,Ey,U);

	U->r=U->r*(double)Nhx*(double)Nhy+(double)Ngx*(double)Ngy*U->i;
	U->i=0.0;
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=U+Ly1;
	     u<v;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--) {
	    t=ey->r/ey->i; r=Ey[hy].r;
			   i=Ey[hy].i; d=(double)Nhx*(r+i*t);
				       g=Product(u->r+v->r,u->i-v->i,d*r, d*i);
			   r=Ey[gy].r;
			   i=Ey[gy].i; d=(double)Ngx*(r+i*t);
				       h=Product(d*r,-d*i,u->i+v->i,u->r-v->r);
	    v->r=u->r=(g.r+h.r)/2.0;
		 u->i=(g.i+h.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*(double)Nhx*(double)Nhy2+(double)Ngx*(double)Ngy2*u->i;
	u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][0]=(*(u++));
/*
    x=1~Lx/2-1
*/
	for (ex=Ex+1, gx=Ngx1, hx=Nhx1, x=1, Lxx=Lx1;
	     x<Lxx;
	     ex++, gx=(gx+Ngx1)%Lx2, hx=(hx+Nhx1)%Lx2, x++, Lxx--) {
	    for (u=U, v=V, y=0; y<Ly; y++) {
		*(u++)=Q[y][x]; *(v++)=Q[y][Lxx];
	    }
	    FFT(-1,Ly,Ey,U); FFT(-1,Ly,Ey,V);

	    t=ex->r/ex->i; r=Ex[hx].r; i=Ex[hx].i; d=r+i*t; hr=d*r; hi=d*i;
			   r=Ex[gx].r; i=Ex[gx].i; d=r+i*t; gr=d*r; gi=d*i;

	    g=Product(U->r+V->r,U->i-V->i,hr*(double)Nhy, hi*(double)Nhy);
	    h=Product(gr*(double)Ngy,-gi*(double)Ngy,U->i+V->i,U->r-V->r);
	    V->r=U->r=(g.r+h.r)/2.0;
		 U->i=(g.i+h.i)/2.0; V->i=(-U->i);

	    for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=V+Ly1, y=1;
		 y<Ly;
		 ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--, y++) {
		t=ey->r/ey->i;
		r=Ey[hy].r;
		i=Ey[hy].i; d=r+i*t; c=Product(hr,hi,d*r,d*i);
				     g=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		r=Ey[gy].r;
		i=Ey[gy].i; d=r+i*t; c=Product(gr,gi,d*r,d*i);
				     h=Product(c.r,-c.i,u->i+v->i,u->r-v->r);
		v->r=u->r=(g.r+h.r)/2.0;
		     u->i=(g.i+h.i)/2.0; v->i=(-u->i);
	    }
	    FFT(1,Ly,Ey,U); FFT(1,Ly,Ey,V);

	    for (u=U, v=V, y=0; y<Ly; y++) {
		Q[y][x]=(*(u++)); Q[y][Lxx]=(*(v++));
	    }
	}
/*
    x=Lx/2
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][x];

	FFT(-1,Ly,Ey,U);

	U->r=U->r*(double)Nhx2*(double)Nhy+(double)Ngx2*(double)Ngy*U->i;
	U->i=0.0;
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=U+Ly1;
	     u<v;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--) {
	    t=ey->r/ey->i;
	    r=Ey[hy].r;
	    i=Ey[hy].i; d=(double)Nhx2*(r+i*t);
			g=Product(u->r+v->r,u->i-v->i,d*r, d*i);
	    r=Ey[gy].r;
	    i=Ey[gy].i; d=(double)Ngx2*(r+i*t);
			h=Product(d*r,-d*i,u->i+v->i,u->r-v->r);
	    v->r=u->r=(g.r+h.r)/2.0;
		 u->i=(g.i+h.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*(double)Nhx2*(double)Nhy2+(double)Ngx2*(double)Ngy2*u->i;
	u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][x]=(*(u++));
/*
    ----
*/
	for (p=P, y=Dy1; y<=Dy2; y++) {
	    FFT(1,Lx,Ex,q=Q[(y<0)?Ly+y:y]);

	    for (x=Dx1; x<=Dx2; x++) *(p++)=q[(x<0)?Lx+x:x].r;
	}
/*
    ----
*/
	for (Gy=G+Ogy, Hy=H+Ohy, y=0; y<Ly; y++) {
	    x=0; q=Q[y];
	    if (y<Ngy)
		for (Gx=(*(Gy++))+Ogx, x=0; x<Ngx; x++) (q++)->r=(*(Gx++));

	    for (; x<Lx; x++) (q++)->r=0.0;

	    x=0; q=Q[y];
	    if (y<Nhy)
		for (Hx=(*(Hy++))+Ohx, x=0; x<Nhx; x++) (q++)->i=(*(Hx++));

	    for (; x<Lx; x++) (q++)->i=0.0;

	    if (y<Ngy || y<Nhy) FFT(-1,Lx,Ex,Q[y]);
	}
/*
    x=0
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][0];

	FFT(-1,Ly,Ey,U);

	U->r*=U->i; U->i=0.0;
	for (u=U+1, v=U+Ly1; u<v; u++, v--) {
	    *u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
		       (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							 v->i=(-u->i);
	}
	u->r*=u->i; u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][0]=(*(u++));
/*
    x=1~Lx/2-1
*/
	for (x=1, Lxx=Lx1; x<Lxx; x++, Lxx--) {
	    for (u=U, v=V, y=0; y<Ly; y++) {
		*(u++)=Q[y][x]; *(v++)=Q[y][Lxx];
	    }
	    FFT(-1,Ly,Ey,U); FFT(-1,Ly,Ey,V);

	    *U=Product((U->r+V->r)/2.0,(U->i-V->i)/2.0,
		       (U->i+V->i)/2.0,(U->r-V->r)/2.0); V->r=  U->r ;
							 V->i=(-U->i);
	    for (u=U+1, v=V+Ly1, y=1; y<Ly; y++, u++, v--) {
		*u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
			   (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							     v->i=(-u->i);
	    }
	    FFT(1,Ly,Ey,U); FFT(1,Ly,Ey,V);

	    for (u=U, v=V, y=0; y<Ly; y++) {
		Q[y][x]=(*(u++)); Q[y][Lxx]=(*(v++));
	    }
	}
/*
    x=Lx/2
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][x];

	FFT(-1,Ly,Ey,U);

	U->r*=U->i; U->i=0.0;
	for (u=U+1, v=U+Ly1; u<v; u++, v--) {
	    *u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
		       (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							 v->i=(-u->i);
	}
	u->r*=u->i; u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][x]=(*(u++));
/*
    ----
*/
	Lxy=(double)Lx*(double)Ly;
	for (p=P, y=Dy1; y<=Dy2; y++) {
	    if (y<0) {
		q=Q[Ly+y]; Lxyy=Lxy*(double)(((z=Nhy+y)<Ngy)?z:Ngy);
	    }
	    else {
		q=Q[   y]; Lxyy=Lxy*(double)(((z=Ngy-y)<Nhy)?z:Nhy);
	    }
	    FFT(1,Lx,Ex,q);

	    for (x=Dx1; x<=Dx2; x++, p++) {
		if (x<0)
		    d=(*p-2.0*q[Lx+x].r)/(double)(((z=Nhx+x)<Ngx)?z:Ngx);
		else
		    d=(*p-2.0*q[   x].r)/(double)(((z=Ngx-x)<Nhx)?z:Nhx);

		*p=(d>0.0)?sqrt(d/Lxyy):0.0;
	    }
	}
/*
    ----
*/
	free(V); free(U); free(*Q); free(Q); free(Ey); free(Ex); return P;
}

int	main(int argc,char **argv)
{
	double	Rx,Ry,*T=NULL;
	char	**path;
	int	Nz,Nx,Ny,x,y,z,Nt,Nl,Nr,y1,y2,x1,x2,Dx,Dx21,Dy,Dy21,dx,dy;
	FOM	**G,**H,f,**F,*D,*d;

	switch(argc) {
	    case  8: case  9: Rx=atof(argv[6]); Ry=atof(argv[7]); break;
	    case 10: case 11: Rx=atof(argv[8]); Ry=atof(argv[9]); break;
	    default:
		Error("usage : "
#ifdef	XR
		      "rr2rc RR/"
#else
		      "ri2rc RI/"
#endif
				" NF top left right {y1 y2} Rx Ry {RMSD.tif}");
	}
	if (Rx<=0.0 || Rx>=1.0 ||
	    Ry<=0.0 || Ry>=1.0) Error("bad distance ratio.");

	path=CheckSliceImages_Float(argv[1],argv[2],NULL,NULL,&Nz,NULL);
	if (Nz<2) Error("too few views.");

	ReadImageFile_Raw(path[   0],&Nx,&Ny,&G);
	ReadImageFile_Raw(path[Nz-1],&x ,&y ,&H);
	if (Nx!=x || Ny!=y) Error("image size not match.");

	for (z=0; z<Nz; z++) free(path[z]); free(path);

	if ((Nt=atoi(argv[3]))>=Ny) Error("bad top margin.");

	if ((Nl=atoi(argv[4]))<0) Nl=0;
	if ((Nr=atoi(argv[5]))<0) Nr=0;
	if (Nl+Nr>=Nx) Error("bad side margin.");

	if (argc==8 || argc==9) {
	    y1=0; y2=Ny-1;
	}
	else
	    if ((y1=atoi(argv[6]))<0 ||
		(y2=atoi(argv[7]))>=Ny || y1>y2) Error("bad slice range.");

	if (Nt>0 && (T=(double *)malloc(sizeof(double)*Nx))==NULL)
	    Error("no memory for I0 correction.");

	I0Correction(Nx,y1,y2,G,Nt,T,Nl,Nr);
	I0Correction(Nx,y1,y2,H,Nt,T,Nl,Nr); if (Nt>0) free(T);

	for (y=y1; y<=y2; y++) {
#ifdef	XR
	    for (x=0; x<Nx; x++) H[y][x]=(-H[y][x]);
#endif
	    for (x1=0, x2=Nx-1; x1<x2; x1++, x2--) {
		f=H[y][x1]; H[y][x1]=H[y][x2]; H[y][x2]=f;
	    }
	}
	Dx21=(Dx=(int)(Rx*(double)Nx          ))*2+1;
	Dy21=(Dy=(int)(Ry*(double)(y2-=(y1-1))))*2+1;

	if (((argc==9 || argc==11) &&
	     (F=(FOM **)malloc(sizeof(FOM *)*Dy21))==NULL) ||
	    (D=RMSD(Nx,y2,G+y1,Nx,y2,H+y1,-Dx,-Dy,Dx,Dy))==NULL)
	    Error("no memory for RMSD calculation.");

	f=(*D); dx=(-Dx); dy=(-Dy);
	for (d=D, y=(-Dy); y<=Dy; y++)
	for (x=(-Dx); x<=Dx; x++, d++)
	    if (f>(*d)) {
		f=(*d); dx=x; dy=y;
	    }

        (void)printf("%lg\t%lg\t%lg\n"
                     "%d\t%d\t%d\n",(double)(Nx-1-Dx)/2.0,
                                    (double)(Nx-1+Dx)/2.0,
                                    (double)(Nx-1+dx)/2.0,-Dy,Dy,dy);
#ifdef	SPR
	(void)printf("%le",f);

	if (dx>(-Dx) && dx<Dx && dy>(-Dy) && dy<Dy)
{
	size_t	i=(size_t)Dx21*(Dy+dy)+Dx+dx;
	double	w=D[i-1],e=D[i+1],we=w-e,wef=we/(w+e-2.0*f)/2.0,
		s=D[i-Dx21],n=D[i+Dx21],sn=s-n,snf=sn/(s+n-2.0*f)/2.0;

	(void)printf("\t%lf\t%lf\t%le",
	((double)(Nx-1+dx)+wef)/2.0,(double)dy+snf,f-(we*wef+sn*snf)/4.0);
}
	(void)putchar('\n');
#endif
	if (argc==9 || argc==11) {
	    F[0]=D; for (y=1; y<Dy21; y++) F[y]=F[y-1]+Dx21;

	    StoreImageFile_Float(argv[argc-1],Dx21,Dy21,F,SIF_F_desc);

	    free(F);
	}
	free(D);

	for (y=0; y<Ny; y++) {
	    free(H[y]); free(G[y]);
	}
	free(H); free(G);

	return 0;
}
