
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "csi_f.h"
#include "rif_r.h"
#include "cbp.h"
#include "sif_f.h"

extern void	Error(char *msg),
		I0Correction(int Nx,int y1,int y2,FOM **I,
			     int Nt,double *T,int Nl,int Nr);

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)(A),0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)(A))
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

#define INIT_MT(T,F,A)	\
if (INIT_THREAD(T,F,A)) Error("multi-threading initialization error.")
#define TERM_MT(T)	\
if (TERM_THREAD(T)) Error("multi-threading termination error.")

#define LEN	2048

typedef struct {
		int	Nx,Ny;
		FOM	**fom;
		char	path[LEN],info[LEN];
	} Struct;

static FUNCTION_T	Store(void *a)
{
	Struct	*s=(Struct *)a;

	StoreImageFile_Float(s->path,s->Nx,s->Ny,s->fom,SIF_F_desc);
	(void)printf("%s\t%s\n",s->info,SIF_F_desc);

	return RETURN_VALUE;
}

#define EPS	1e-9
#define DEG	3.14159265358979323846/180.0

int	main(int argc,char **argv)
{
	char		**path,form[LEN];
	int		Nx,Ny,Nv,Nt,Nl,Nr,layer,count,Nz,v,x,y,z,len,c;
	double		Dr,base,step,RA0,y0,dy,RC,Dz=0.0,*top=NULL;
	Float		**P,*P_F,**F;
	FOM		**I,*s,*i;
	Struct		S;
	THREAD_T	T;

#ifdef	XR
	if (argc!=14 && argc!=15)
	    Error(
"usage :\n"
"rr_stg RR/ NF top left right layer SDD {-}Dp RC_base step count {Dz} RA0 STG/"
	    );
#else
	if (argc!=13 && argc!=14)
	    Error(
"usage : ri_stg RI/ NF top left right layer Dr RC_base step count {Dz} RA0 STG/"
	    );
#endif
	path=CheckSliceImages_Float(argv[1],argv[2],&Nx,&Ny,&Nv,NULL);

	if (Nv<1) Error("too few views.");

	free(path[--Nv]);

	if ((Nt=atoi(argv[3]))>=Ny) Error("bad top margin.");

	if ((Nl=atoi(argv[4]))<0) Nl=0;
	if ((Nr=atoi(argv[5]))<0) Nr=0;
	if (Nl+Nr>=Nx) Error("bad side margin.");

	layer=atoi(argv[6]);

#ifdef	XR
{
	double	SDD=atof(argv[7]),
		Dp=atof(argv[8]);

	if (SDD<EPS) Error("bad sample-detector distance.");

	if (-EPS<Dp && Dp<EPS)
	    Error("bad interval of detectors on refraction profile.");

	Dr=SDD/Dp;
}

#define NO_DZ	14
#else
	if ((Dr=atof(argv[7]))<EPS)
	    Error("bad horizontal interval of detectors on raw image.");

#define NO_DZ	13
#endif
	if ((base=atof(argv[NO_DZ-5]))<=0.0 ||
	    (step=atof(argv[NO_DZ-4]))<=0.0 ||
	    (count=atoi(argv[NO_DZ-3]))<=0 ||
	    base+step*(double)(count-1)>=(double)(Nx-1))
	    Error("bad range of rotation center.");

	if (argc==NO_DZ)
	    Nz=Ny;
	else
	    if ((Nz=Ny+(int)ceil(fabs(Dz=atof(argv[NO_DZ-2]))))>=2*Ny)
		Error("too large vertical displacement.");

        if (layer<0 || layer>=Nz) Error("bad slice position.");

	RA0=atof(argv[argc-2])*DEG;

	P=InitCBP(Nx,Nv);

	if ((S.fom=(FOM **)malloc(sizeof(FOM *)*(v=(Nx>Nv)?Nx:Nv)))==NULL ||
	    (*(S.fom)=(FOM *)malloc(sizeof(FOM)*v*Nx))==NULL ||
	    (Nt>0 && (top=(double *)malloc(sizeof(double)*Nx))==NULL))
	    Error("memory allocation error.");

	for (x=1; x<v; x++) S.fom[x]=S.fom[x-1]+Nx;

	y0=0.5+(double)layer; if (Dz<0.0) y0+=Dz; dy=Dz/(double)Nx;
	for (v=0; v<Nv; v++) {
	    P_F=P[v]; s=S.fom[v];
	    if ((y=(int)floor(y0-dy*(double)v))<0 || y>=Ny)
		for (x=0; x<Nx; x++) *(P_F++)=(*(s++))=0.0;
	    else {
		ReadImageFile_Raw(path[v],&x,&z,&I);
		if (Nx!=x || Ny!=z) Error("image size not match.");

		I0Correction(Nx,y,y,I,Nt,top,Nl,Nr);

		i=I[y]; for (x=0; x<Nx; x++) *(P_F++)=(*(s++))=(*(i++));

		for (y=0; y<Ny; y++) free(I[y]); free(I);
	    }
	    free(path[v]);
	}
	free(path); if (Nt>0) free(top);

	len=1; for (z=10; z<Nz; z*=10) ++len;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],len);
	(void)sprintf(S.path,form,layer);
	(void)sprintf(S.info,"%d",layer);

	S.Nx=Nx; S.Ny=Nv;

	INIT_MT(T,Store,&S);

	c=1+2; for (x=10; x<Nx-1; x*=10) ++c;
	(void)sprintf(form,"%s/%%0%dd_%%0%d.1lf.tif",argv[argc-1],len,c);

	for (c=0; c<count; c++) {
	    F=CBP(Dr,-(RC=base+step*(double)c),RA0);

	    TERM_MT(T);

	    if (c==0) S.Ny=Nx;

	    for (y=0; y<Nx; y++) {
		s=S.fom[y]; P_F=F[y]; for (x=0; x<Nx; x++) *(s++)=(*(P_F++));
	    }
	    (void)sprintf(S.path,form,layer,RC);
	    (void)sprintf(S.info,"%d\t%lf",layer,RC);

	    INIT_MT(T,Store,&S);
	}
	TERM_MT(T); free(*(S.fom)); free(S.fom); TermCBP();

	return 0;
}
